/* Various display routines and vector/matrix mainpulation routines */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <fcntl.h>
#ifdef TURBOC
#include <alloc.h>
#include <dos.h>
#include <graphics.h>
#endif
#include "wirefram.h"

#ifndef M_PI
#define M_PI 3.1415926535897932384626
#endif

/* The definitions to glue between different sets of manipulation
   functions for vectors */
#define MakeVec(a, b, c, v)      (v).x=(a),(v).y=(b),(v).z=(c)
#define VecScale(S,a)   (a).x *= S ; (a).y *= S ; (a).z *= S
#define VecNegate(a)   (a).x=0.0-(a).x;(a).y=0.0-(a).y;(a).z=0.0-(a).z;
#define VecDot(a,b)   ((a).x*(b).x+(a).y*(b).y+(a).z*(b).z)
#define VecLen(a)   (sqrt(VecDot(a,a)))
#define VecCopy(a,b)    (b).x=(a).x;(b).y=(a).y;(b).z=(a).z;
#define VecAdd(a,b,c) (c).x=(a).x+(b).x;(c).y=(a).y+(b).y;(c).z=(a).z+(b).z
#define VecSub(a,b,c) (c).x=(a).x-(b).x;(c).y=(a).y-(b).y;(c).z=(a).z-(b).z
#define VecCross(a,b,c)    (c).x=(a).y*(b).z-(a).z*(b).y;\
          (c).y=(a).z*(b).x-(a).x*(b).z;\
          (c).z=(a).x*(b).y-(a).y*(b).x
typedef struct {
   VECTOR from; /* Position of the eye */
   VECTOR at;   /* Point being looked at */
   VECTOR up;   /* Up direction */
   VECTOR right;
   } Eye_Struct;

int XResolution = 160;
int YResolution = 160;
static Eye_Struct NewEye, OldEye;
static DBL Focal_Distance = 5.0;
static DBL Aspect = 1.0;
static DBL Field_Of_View = 1.0;

extern Depth_Map_Type Depth_Map[64];
extern Screen_Map_Type Screen_Map[64];
extern unsigned long Total_Map_Lines, Total_Map_Buffers, Total_Elements;
extern unsigned long First_Map_Line, Last_Map_Line;
extern unsigned long Total_Lines_Per_Buffer, Total_Elements_Per_Buffer;

extern int window_width, window_height;
extern double offsetx, offsety;
extern int VGA_Flag;

extern unsigned long int PATTERN; /* Pattern to use in line draws */
extern int LINEWIDTH;             /* Width of lines */

extern DBL VTemp;
extern OBJECT *Light_Sources;
extern VIEWPOINT *Eye;
extern TRANSFORMATION WS;

void
put_depth(x, y, depth)
   int x, y;
   depth_type depth;
{
   unsigned long offset;
   int buffer, line;

   if (y >= First_Map_Line && y <= Last_Map_Line) {
      y -= First_Map_Line;
      buffer = y / Total_Lines_Per_Buffer;
      line   = y % Total_Lines_Per_Buffer;
      offset = (long)((long)window_width * (long)line) + x;
      Depth_Map[buffer][offset] = depth;
      }
}

void
put_color(int x, int y, DBL red, DBL green, DBL blue)
{
   unsigned long offset;
   int buffer, line;

   if (y >= First_Map_Line && y <= Last_Map_Line) {
      y -= First_Map_Line;
      buffer = y / Total_Lines_Per_Buffer;
      line   = y % Total_Lines_Per_Buffer;
      offset = (long)((long)window_width * (long)line) + x;
      Screen_Map[buffer][offset].red   = (unsigned char)(255.0 * red);
      Screen_Map[buffer][offset].green = (unsigned char)(255.0 * green);
      Screen_Map[buffer][offset].blue  = (unsigned char)(255.0 * blue);
      }
}

depth_type
get_depth(x, y)
   int x, y;
{
   long offset;
   int buffer, line;

   if (y >= First_Map_Line && y <= Last_Map_Line) {
      y -= First_Map_Line;
      buffer = y / Total_Lines_Per_Buffer;
      line   = y % Total_Lines_Per_Buffer;
      offset = (long)((long)window_width * (long)line) + x;
      return Depth_Map[buffer][offset];
      }
   else
      return 0;
}

void
show_polygon(color, Close_Flag, vertex_count, x, y)
   TEXTURE *color;
   int Close_Flag, vertex_count;
   DBL *x, *y;
{
   int i;
   VECTOR c;

   c.x = color->Color[0];
   c.y = color->Color[1];
   c.z = color->Color[2];
   /* Call the routine to draw the triangle */
   setcolor(get_indexed_color(c));
   for (i=0;i<vertex_count;i++)
      convert(x[i], y[i]);
   for (i=0;i<vertex_count-1;i++)
      line(x[i], y[i], x[i+1], y[i+1]);
   if (Close_Flag)
      line(x[vertex_count-1], y[vertex_count-1], x[0], y[0]);
}

void
show_rectangle(color, v0, v1, v2, v3)
   TEXTURE *color;
   VECTOR *v0, *v1, *v2, *v3;
{
   VECTOR v[4];
   DBL vx[4], vy[4];
   int i, shade;
   VECTOR c;

   c.x = color->Color[0];
   c.y = color->Color[1];
   c.z = color->Color[2];

   /* Transform based on distance from eye and placement on screen */
   v[0] = *v0; v[1] = *v1; v[2] = *v2; v[3] = *v3;
   project_points(Eye, 4, &v[0], &vx[0], &vy[0], NULL);
   for (i=0;i<4;i++) convert(vx[i], vy[i]);
   setcolor(get_indexed_color(c));
   line(vx[0], vy[0], vx[1], vy[1]);
   line(vx[1], vy[1], vx[2], vy[2]);
   line(vx[2], vy[2], vx[3], vy[3]);
   line(vx[3], vy[3], vx[0], vy[0]);
}

void
drawRect(x0, y0, x1, y1, color)
   int x0, y0, x1, y1, color;
{
   setcolor(color);
   line(x0, y0, x0, y1);
   line(x0, y1, x1, y1);
   line(x1, y1, x1, y0);
   line(x1, y0, x0, y0);
}

/* Calculate the normal to a polygon.  1 is returned if there are enough
   distinct vertices on the polygon to find a common plane.  0 is returned
   if the polygon is degenerate. */
int
polygon_normal(vertex_count, vertices, Result, d)
   int vertex_count;
   VECTOR *vertices, *Result;
   DBL *d;
{
   VECTOR verts[6];
   VECTOR V1, V2;
   DBL Length;
   int i, j, vcnt=0;
   DBL testval;

   /* Grab the first three non-equal vertices */
   for (i=0;i<vertex_count && vcnt < 3;i++) {
      for (j=0;j<vcnt;j++)
         if ((fabs(vertices[i].x - vertices[j].x) < EPSILON) &&
             (fabs(vertices[i].y - vertices[j].y) < EPSILON) &&
             (fabs(vertices[i].z - vertices[j].z) < EPSILON))
            goto l0;
      verts[vcnt++] = vertices[i];
      l0:;
      }

   if (vcnt < 3) {
      VSet(*Result, 1.0, 0.0, 0.0);
      *d = 0.0;
      printf("Too few vertices in polygon_normal: %d\n", vcnt);
      return 0;
      }

   VSub(V1, verts[1], verts[0]);
   VSub(V2, verts[2], verts[0]);
   VCross(*Result, V1, V2);
   VLength(Length, *Result);
   if (Length < EPSILON) {
      VSet(*Result, 1.0, 0.0, 0.0);
      *d = 0.0;
      /* printf("Bad normal in polygon_normal\n"); */
      return 0;
      }
   else {
      VInverseScale(*Result, *Result, Length);
      VDot(*d, *Result, verts[0]);
      *d = 0.0 - *d;
      return 1;
      }
}

/* Calculate the various plane equations for a polygon.  1 is returned if
   there are enough distinct vertices on the polygon to find a common
   plane.  0 is returned if the polygon is degenerate.

   The inputs are:
      vertex_count -    The total # of vertices in the polygon (>= 3)
      vertices     -    A pointer to an array of vectors

   The values returned from this routine are:
      n  -      The normal to the polygon's plane
      d  -      The value d in the plane equation ax + by + cz + d = 0
      l0 -      The length of the first basis vector of the plane
      l1 -      The length of the second basis vector of the plane
      b0 -      The first basis vector for the plane (normalized)
      b1 -      The second basis vector for the plane (normalized)
*/
int
determine_plane_equations(vertex_count, vertices, n, d, l0, l1, b0, b1)
   int vertex_count;
   VECTOR *vertices, *n;
   DBL *d, *l0, *l1;
   VECTOR *b0, *b1;
{
   int i;
   VECTOR d0, d1, verts[3];
   DBL temp;

   /* Grab the three non-equal vertices. Try to get them so that they
      determine a basis that includes the polygon within the positive
      direction along each basis vector. */
   verts[0] = vertices[0];
   for (i=1;i<vertex_count;i++)
      if ((fabs(vertices[i].x - verts[0].x) > EPSILON) ||
          (fabs(vertices[i].y - verts[0].y) > EPSILON) ||
          (fabs(vertices[i].z - verts[0].z) > EPSILON)) {
         verts[1] = vertices[i];
         break;
         }
   if (i == vertex_count) {
      printf("Failed to find vert[1]\n");
      return 0;
      }
   for (i=vertex_count-1;i>0;i--)
      if (((fabs(vertices[i].x - verts[0].x) > EPSILON) ||
           (fabs(vertices[i].y - verts[0].y) > EPSILON) ||
           (fabs(vertices[i].z - verts[0].z) > EPSILON)) &&
          ((fabs(vertices[i].x - verts[1].x) > EPSILON) ||
           (fabs(vertices[i].y - verts[1].y) > EPSILON) ||
           (fabs(vertices[i].z - verts[1].z) > EPSILON))) {
         verts[2] = vertices[i];
         break;
         }
   if (i == 0) {
      printf("Failed to find vert[2]\n");
      return 0;
      }

   /* Determine first basis vector */
   VSub(d0, verts[1], verts[0]);
   *l0 = sqrt(d0.x * d0.x + d0.y * d0.y + d0.z * d0.z);
   VInverseScale(*b0, d0, *l0);

   /* Determine second basis vector */
   VSub(d1, verts[2], verts[0]);
   *l1 = sqrt(d1.x * d1.x + d1.y * d1.y + d1.z * d1.z);
   VInverseScale(*b1, d1, *l1);

   /* Determine plane equations */
   VCross(*n, *b0, *b1);
   VNormalize(*n, *n);
   VDot(temp, *n, verts[0]);
   *d = 0.0 - temp;
   return 1;
}

/* Solve a set of linear equations to determine how a point in a plane
   is related to the basis vectors for the plane. */
int
find_basis_values(u0, u1, p0, a, b)
   VECTOR *u0, *u1, *p0;
   DBL *a, *b;
{
   DBL x0, y0, x1, y1, x2, y2;
   if (fabs(u0->x) < EPSILON && fabs(u1->x) < EPSILON)
      if (fabs(u0->y) < EPSILON && fabs(u1->y) < EPSILON)
         return 0;
      else if (fabs(u0->z) < EPSILON && fabs(u1->z) < EPSILON)
         return 0;
      else {
         x0 = u0->y; x1 = u1->y; x2 = p0->y;
         y0 = u0->z; y1 = u1->z; y2 = p0->z;
         }
   else if (fabs(u0->y) < EPSILON && fabs(u1->y) < EPSILON)
      if (fabs(u0->z) < EPSILON && fabs(u1->z) < EPSILON)
         return 0;
      else {
         x0 = u0->x; x1 = u1->x; x2 = p0->x;
         y0 = u0->z; y1 = u1->z; y2 = p0->z;
         }
   else {
      x0 = u0->x; x1 = u1->x; x2 = p0->x;
      y0 = u0->y; y1 = u1->y; y2 = p0->y;
      }
   if (fabs(x0) < EPSILON) {
      /* Solve directly for b */
      *b = x2 / x1;
      *a = (y2 - *b * y1) / y0;
      }
   else if (fabs(x1) < EPSILON) {
      /* Solve directly for a */
      *a = x2 / x0;
      *b = (y2 - *a * y0) / y1;
      }
   else if (fabs(y0) < EPSILON) {
      /* Solve directly for b */
      *b = y2 / y1;
      *a = (x2 - *b * x1) / x0;
      }
   else if (fabs(y1) < EPSILON) {
      /* Solve directly for a */
      *a = y2 / y0;
      *b = (x2 - *a * x0) / x1;
      }
   else {
      *b = (x0 * y2 - x2 * y0) / (x0 * y1 - x1 * y0);
      *a = (x2 - *b * x1) / x0;
      }
   return 1;
}

/* Determine the distance from a point to a plane. */
DBL
point_plane_distance(p, n, d)
   VECTOR *p, *n;
   DBL *d;
{
   DBL temp1, temp2;

   VDot(temp1, *p, *n);
   temp1 += *d;
   VLength(temp2, *n);
   if (fabs(temp2) < EPSILON) return 0;
   temp1 /= temp2;
   return temp1;
}

/* Rotate a list of points about the X, Y, and Z axes (in that order) */
void
rotate_points(Angles, numpoints, rot_pts, pt)
   int numpoints;
   VECTOR *Angles;
   VECTOR *rot_pts, *pt;
{
   int i,j;
   DBL cx, sx, cy, sy, cz, sz;
   VECTOR v0;
   DBL x_angle = Angles->x;
   DBL y_angle = Angles->y;
   DBL z_angle = Angles->z;

   cx = cos(x_angle); sx = sin(x_angle);
   cy = cos(y_angle); sy = sin(y_angle);
   cz = cos(z_angle); sz = sin(z_angle);

   for (i=0; i<numpoints; i++) {
      v0.x = pt[i].z * cz * sy * cx -
             pt[i].y * sz * cx +
             pt[i].y * cz * sy * sx +
             pt[i].z * sz * sx +
             pt[i].x * cz * cy;

      v0.y = pt[i].z * sz * sy * cx +
             pt[i].y * cz * cx +
             pt[i].y * sz * sy * sx -
             pt[i].z * cz * sx +
             pt[i].x * sz * cy;

      v0.z = pt[i].z * cy * cx +
             pt[i].y * cy * sx -
             pt[i].x * sy;

      rot_pts[i] = v0;
      }
}

/* Return the coefficients a..d of the plane equation aX + bY + cZ + d = 0
   for the plane that is common to the three points x, y, and z.  If the
   test point "t" is non-NULL then the normal will be forced to point
   away from "t". */
int
common_plane(VECTOR *x, VECTOR *y, VECTOR *z, VECTOR *t, VECTOR *normal, DBL *d)
{
  /* Allocate storage for intermediate results. */
  VECTOR v1, v2, v3;
  DBL temp, testval;

  VSub(v1, *y, *x);
  VSub(v2, *z, *x);
  VCross(v3, v1, v2);
  VLength(temp, v3);
  if (temp == 0.0) {
     /* The points are probably co-linear, couldn't find a common plane. */
     VSet(*normal, 0.0, 0.0, 0.0);
     *d = 0.0;
     return 0;
     }
  VInverseScale(v3, v3, temp);
  *normal = v3;
  VDot(temp, v3, *x);
  *d = -temp;

  /* Make sure the normal points the right way. */
  if (t != NULL) {
     VDot(temp, v3, *t);
     temp += *d;
     if (temp > 0.0) {
        /* Flip the normal. */
        VScale(*normal, v3, -1.0);
        *d *= -1.0;
        }
     }
  return 1;
}

/* Make sure the point "t" is below the plane described by "n" and "d".  This
   is is true if the solution of the plane equation is negative when the
   point "t" is substituted. */
void
point_normal(n, d, t)
   VECTOR *n, *t;
   DBL *d;
{
  DBL temp;

  /* Make sure the normal points the right way. */
  if (t != NULL) {
     VDot(temp, *n, *t);
     temp += *d;
     if (temp > 0.0) {
        /* Flip the normal. */
        VScale(*n, *n, -1.0);
        *d *= -1.0;
        }
     }
}

DBL
degrees_to_radians(DBL degrees)
{
   DBL angle;
   while (degrees >= 360)
      degrees -= 360;
   while (degrees < 0)
      degrees += 360;
   angle = rad_per_degree * degrees;
   return angle;
}

DBL
radians_to_degrees(DBL radians)
{
   DBL angle;
   if (radians >= M_PI) radians -= 2.0 * M_PI;
   if (radians <= -M_PI) radians += 2.0 * M_PI;
   angle = degree_per_rad * radians;
   if (angle < -180  || angle > 180) {
      printf("Angle too big: %f\n", angle);
      }
   return angle;
}

/* Evaluate a single coordinate point (u, v) on a bezier patch. */
void
bezier_value(Result, u, v, Patch)
   VECTOR *Result;
   DBL u, v;
   VECTOR (*Patch)[4][4];
{
   DBL u2, u3, v2, v3, uu1, uu2, uu3, vv1, vv2, vv3;
   DBL t[4][4];
   int i, j;

   u2 = u * u; u3 = u * u2; v2 = v * v; v3 = v * v2;
   uu1 = 1.0 - u; uu2 = uu1 * uu1; uu3 = uu1 * uu2;
   vv1 = 1.0 - v; vv2 = vv1 * vv1; vv3 = vv1 * vv2;
   t[0][0] =             uu3 *      vv3;
   t[0][1] = 3.0 *       uu3 * v  * vv2;
   t[0][2] = 3.0 *       uu3 * v2 * vv1;
   t[0][3] =             uu3 * v3;
   t[1][0] = 3.0 *  u  * uu2 *      vv3;
   t[1][1] = 9.0 *  u  * uu2 * v  * vv2;
   t[1][2] = 9.0 *  u  * uu2 * v2 * vv1;
   t[1][3] = 3.0 *  u  * uu2 * v3;
   t[2][0] = 3.0 *  u2 * uu1 *      vv3;
   t[2][1] = 9.0 *  u2 * uu1 * v  * vv2;
   t[2][2] = 9.0 *  u2 * uu1 * v2 * vv1;
   t[2][3] = 3.0 *  u2 * uu1 * v3;
   t[3][0] =        u3 *            vv3;
   t[3][1] = 3.0 *  u3 *       v  * vv2;
   t[3][2] = 3.0 *  u3 *       v2 * vv1;
   t[3][3] =        u3 *       v3;

   Result->x = 0.0;
   Result->y = 0.0;
   Result->z = 0.0;
   for (i=0;i<4;i++) for (j=0;j<4;j++) {
      Result->x += t[i][j] * (*Patch)[i][j].x;
      Result->y += t[i][j] * (*Patch)[i][j].y;
      Result->z += t[i][j] * (*Patch)[i][j].z;
      }
}

void
bezier_partial_u(Result, u, v, Patch)
   DBL u, v;
   VECTOR *Result, (*Patch)[4][4];
{
   DBL u2, v2, v3;
   DBL t[4][4];
   int i, j;
   u2 = u * u; v2 = v * v; v3 = v * v2;

   t[0][0] = 3.0*(v3-3.0*v2+3.0*v-1.0)*(u2-2.0*u+1.0);
   t[0][1] = 9.0*v*(v2-2.0*v+1.0)*(u2-2.0*u+1.0);
   t[0][2] = 9.0*v2*(v-1.0)*(u2-2.0*u+1.0);
   t[0][3] = 3.0*v3*(u2-2.0*u+1.0);
   t[1][0] = 3.0*(v3-3.0*v2+3.0*v-1)*(3.0*u2-4.0*u+1);
   t[1][1] = 9.0*v*(v2-2.0*v+1.0)*(3.0*u2-4.0*u+1.0);
   t[1][2] = 9.0*v2*(v-1.0)*(3.0*u2-4.0*u+1.0);
   t[1][3] = 3.0*v3*(3.0*u2-4.0*u+1.0);
   t[2][0] = 3.0*u*(v3-3.0*v2+3.0*v-1.0)*(3.0*u-2.0);
   t[2][1] = 9.0*u*v*(v2-2.0*v+1.0)*(3.0*u-2.0);
   t[2][2] = 9.0*u*v2*(v-1.0)*(3.0*u-2.0);
   t[2][3] = 3.0*u*v3*(3.0*u-2.0);
   t[3][0] = 3.0*u2*(v3-3.0*v2+3.0*v-1.0);
   t[3][1] = 9.0*u2*v*(v2-2.0*v+1.0);
   t[3][2] = 9.0*u2*v2*(v-1.0);
   t[3][3] = 3.0*u2*v3;

   VSet(*Result, 0.0, 0.0, 0.0);
   for (i=0;i<4;i++) for (j=0;j<4;j++) {
      Result->x += t[i][j] * (*Patch)[i][j].x;
      Result->y += t[i][j] * (*Patch)[i][j].y;
      Result->z += t[i][j] * (*Patch)[i][j].z;
      }
}

void
bezier_partial_v(Result, u, v, Patch)
   DBL u, v;
   VECTOR *Result, (*Patch)[4][4];
{
   DBL u2, u3, v2;
   DBL t[4][4];
   int i, j;
   u2 = u * u; u3 = u * u2; v2 = v * v;
   t[0][0] = 3.0*(v2-2.0*v+1.0)*(u3-3.0*u2+3.0*u-1.0);
   t[0][1] = 3.0*(3.0*v2-4.0*v+1.0)*(u3-3.0*u2+3.0*u-1.0);
   t[0][2] = 3.0*v*(3.0*v-2.0)*(u3-3.0*u2+3.0*u-1.0);
   t[0][3] = 3.0*v2*(u3-3.0*u2+3.0*u-1.0);
   t[1][0] = 9.0*u*(v2-2.0*v+1.0)*(u2-2.0*u+1.0);
   t[1][1] = 9.0*u*(3.0*v2-4.0*v+1.0)*(u2-2.0*u+1.0);
   t[1][2] = 9.0*u*v*(3.0*v-2.0)*(u2-2.0*u+1.0);
   t[1][3] = 9.0*u*v2*(u2-2.0*u+1.0);
   t[2][0] = 9.0*u2*(v2-2.0*v+1.0)*(u-1.0);
   t[2][1] = 9.0*u2*(3.0*v2-4.0*v+1.0)*(u-1.0);
   t[2][2] = 9.0*u2*v*(3.0*v-2.0)*(u-1.0);
   t[2][3] = 9.0*u2*v2*(u-1.0);
   t[3][0] = 3.0*u3*(v2-2.0*v+1.0);
   t[3][1] = 3.0*u3*(3.0*v2-4.0*v+1.0);
   t[3][2] = 3.0*u3*v*(3.0*v-2.0);
   t[3][3] = 3.0*u3*v2;

   VSet(*Result, 0.0, 0.0, 0.0);
   for (i=0;i<4;i++) for (j=0;j<4;j++) {
      Result->x += t[i][j] * (*Patch)[i][j].x;
      Result->y += t[i][j] * (*Patch)[i][j].y;
      Result->z += t[i][j] * (*Patch)[i][j].z;
      }
}

void
bezier_normal(Result, u, v, Patch)
   DBL u, v;
   VECTOR *Result, (*Patch)[4][4];
{
   DBL len;
   VECTOR u_normal, v_normal;
   bezier_partial_u(&u_normal, u, v, Patch);
   bezier_partial_v(&v_normal, u, v, Patch);
   VCross(*Result, u_normal, v_normal);
   VLength(len, *Result);
   if (fabs(len) < EPSILON) {
      VSet(*Result, 1.0, 0.0, 0.0);
      }
   else
      VInverseScale(*Result, *Result, len);
}

/* Given a set of reference axes, generate the angles needed to rotate
   them into co-incidence with the coordinate axes. Major_Axis -> Z,
   Minor_Axis1 -> Y, Minor_Axis2 -> X. The assumption is made that the
   reference axes are all normalized.  The resulting angles are given
   in radians. */
void
determine_reference_angles(Axes, Angles)
   VECTOR *Axes;
   VECTOR *Angles;
{
   DBL x1, y1, l;
   DBL theta;
   VECTOR Axis1, Temp_Angles;

   /* Rotate the major axis into the Y-Z plane */
   x1 = Axes[0].x;
   y1 = Axes[0].z;
   l  = sqrt(x1 * x1 + y1 * y1);
   if (l != 0.0) {
      theta = acos(y1 / l);
      if (x1 > 0) theta *= -1.0;
      Angles->y = theta;
      }
   else
      Angles->y = 0.0;

   VSet(Temp_Angles, 0.0, theta, 0.0);
   Axis1 = Axes[0];
   rotate_points(&Temp_Angles, 1, &Axis1, &Axis1);

   /* Bring the major axis into alignment with the z-axis */
   x1 = Axis1.y; /* Axes[0].y; */
   y1 = Axis1.z; /* Axes[0].z; */
   l = sqrt(x1 * x1 + y1 * y1);
   if (l != 0.0) {
      theta = acos(y1 / l);
      if (x1 < 0) theta *= -1.0;
      Angles->x = theta;
      }
   else
      Angles->x = 0.0;

   /* Rotate Minor_Axis1 to see how much roll is needed to get everything
      into accord. */
   Axis1 = Axes[1];
   VSet(Temp_Angles, 0.0, Angles->y, 0.0);
   rotate_points(&Temp_Angles, 1, &Axis1, &Axis1);

   VSet(Temp_Angles, Angles->x, 0.0, 0.0);
   rotate_points(&Temp_Angles, 1, &Axis1, &Axis1);

   x1 = Axis1.x;
   y1 = Axis1.y;
   l = sqrt(x1 * x1 + y1 * y1);
   if (l != 0.0) {
      theta = acos(y1 / l);
      if (x1 < 0) theta *= -1.0;
      Angles->z = theta;
      }
   else {
      printf("Bad length in determine_reference_angles\n");
      }
   VSet(*Angles, -Angles->x, -Angles->y, -Angles->z);
}

/* Find the transform that takes the current eye position and
   orientation and puts it at {0, 0, -Focal_Distance} with the up vector aligned
   with {0, 1, 0} */
void
Normalize_View(TRANSFORMATION *Tv)
{
   DBL x, y, z, d, theta;
   DBL xtheta = 0.0, ytheta = 0.0, ztheta = 0.0;
   VECTOR Va, Ve, Vv, Rt;
   DBL XScale = 1.0;
   DBL YScale = 1.0;
   TRANSFORMATION *Tt = Get_Transformation();

   NewEye.from    = Eye->Orientation.Origin;
   NewEye.up      = Eye->Orientation.Minor_Axis1;
   NewEye.right   = Eye->Orientation.Minor_Axis2;
   NewEye.at      = Eye->Focus->Look_At;
   Focal_Distance = Eye->Focus->Distance;
   Field_Of_View  = Eye->Focus->Field_Of_View;
   Aspect         = Eye->Focus->Aspect;
   XResolution    = window_width;
   YResolution    = window_height;
   OldEye         = NewEye;
/*
printf("View from  <%f, %f, %f>, Dist = %f\n",
       NewEye.from.x, NewEye.from.y, NewEye.from.z,
       Focal_Distance);
printf("View right <%f, %f, %f>\n",
       NewEye.right.x, NewEye.right.y, NewEye.right.z);
printf("View up    <%f, %f, %f>\n",
       NewEye.up.x, NewEye.up.y, NewEye.up.z);
printf("At         <%f, %f, %f>\n",
       NewEye.at.x, NewEye.at.y, NewEye.at.z);
*/
   /* Translate point of interest to the origin */
   VecCopy(NewEye.at, Va);

   /* Is the translation coming out backwards? */
   VecNegate(Va);
   Get_Translation_Transformation(Tv, &Va);

   /* Move the eye by the same amount */
   VecAdd(NewEye.from, Va, Ve);

   /* Make sure this is a valid sort of setup */
   if (Ve.x == 0.0 && Ve.y == 0.0 && Ve.z == 0.0) {
      printf("Eye position is same as point of interest\n");
      exit(1);
      }

   /* Determine the angle of rotation about the x-axis that takes
     the eye into the x-z plane. */
   y = Ve.y;
   if (fabs(y) > 1.0e-10) {
      z = Ve.z;
      d = sqrt(y*y + z*z);
      if (fabs(d) > 1.0e-5) {
         if (y > 0)
            theta = acos(z/d) - M_PI;
         else
            theta = -acos(z/d) - M_PI;
         xtheta = theta;
         MakeVec(theta*180.0/M_PI, 0.0, 0.0, Vv);
         Get_Rotation_Transformation(Tt, &Vv);
         MTransformVector(&Vv, &Ve, Tt);
         VecCopy(Vv, Ve);
         Compose_Transformations(Tv, Tt);
         }
      }

   /* Determine the angle of rotation about the y-axis that takes
      the eye onto the minus z-axis */
   MTransformVector(&Ve, &NewEye.from, Tv);
   x = Ve.x;
   if (fabs(x) > 1.0e-10) {
      z = Ve.z;
      d = sqrt(x*x + z*z);
      if (fabs(d) > 1.0e-5) {
         if (x > 0)
            theta = acos(-z/d);
         else
            theta = -acos(-z/d);
         ytheta = theta;
         MakeVec(0.0, theta*180.0/M_PI, 0.0, Vv);
         Get_Rotation_Transformation(Tt, &Vv);
         MTransformVector(&Vv, &Ve, Tt);
         Compose_Transformations(Tv, Tt);
         VecCopy(Vv, Ve);
         }
      }
   else if (Ve.z > 0) {
      MakeVec(0.0, 180.0, 0.0, Vv);
      ytheta = M_PI;
      Get_Rotation_Transformation(Tt, &Vv);
      MTransformVector(&Vv, &Ve, Tt);
      Compose_Transformations(Tv, Tt);
      VecCopy(Vv, Ve);
      }

   /* Determine the angle of rotation about the z-axis that takes
      the up vector into alignment with the y-axis */
   MTransVector(&Ve, &NewEye.up, Tv);
   y = Ve.y;
   if (fabs(y) > 1.0e-10) {
      x = Ve.x;
      d = sqrt(x*x + y*y);
      if (fabs(d) > 1.0e-5) {
         if (x > 0)
            theta = acos(y/d);
         else
            theta = -acos(y/d);
         ztheta = theta;
         MakeVec(0.0, 0.0, theta*180.0/M_PI, Vv);
         Get_Rotation_Transformation(Tt, &Vv);
         MTransformVector(&Vv, &Ve, Tt);
         Compose_Transformations(Tv, Tt);
         VecCopy(Vv, Ve);
         }
      }
   else if (Ve.x > 0) {
      MakeVec(0.0, 0.0, M_PI/2.0, Vv);
      ztheta = M_PI/2.0;
      Get_Rotation_Transformation(Tt, &Vv);
      MTransformVector(&Vv, &Ve, Tt);
      Compose_Transformations(Tv, Tt);
      VecCopy(Vv, Ve);
      }
   else {
      MakeVec(0.0, 0.0, -M_PI/2.0, Vv);
      ztheta = -M_PI/2.0;
      Get_Rotation_Transformation(Tt, &Vv);
      MTransformVector(&Vv, &Ve, Tt);
      Compose_Transformations(Tv, Tt);
      VecCopy(Vv, Ve);
      }

   /* Last, translate the eye position to the point <0, 0, -Focal_Distance> */
   MTransformVector(&Ve, &NewEye.from, Tv);
   MakeVec(0.0, 0.0, -Focal_Distance - Ve.z, Va);
   Get_Translation_Transformation(Tt, &Va);
   Compose_Transformations(Tv, Tt);

   /* Determine how much to scale things by */
   YScale = (DBL)YResolution / (2.0 * tan(Field_Of_View) * Focal_Distance);
   XScale = YScale * (DBL)XResolution / ((DBL)YResolution * Aspect);
   MakeVec(XScale, YScale, 1.0, Va);
   Get_Scaling_Transformation(Tt, &Va);
   Compose_Transformations(Tv, Tt);

/*
printf("Field of view = %f, Dist = %f, XScale = %f, YScale = %f\n",
       2.0 * 180.0 * Field_Of_View / 3.14159265,
       Focal_Distance, XScale, YScale);
*/
   /* Now add in the perspective transformation */
   Get_Perspective_Transformation(Tt, Focal_Distance);
   Compose_Transformations(Tv, Tt);
/*
printf("Transform is:\n");
printf("  <%f, %f, %f, %f>\n",
       Tv->matrix[0][0], Tv->matrix[0][1], Tv->matrix[0][2], Tv->matrix[0][3]);
printf("  <%f, %f, %f, %f>\n",
       Tv->matrix[1][0], Tv->matrix[1][1], Tv->matrix[1][2], Tv->matrix[1][3]);
printf("  <%f, %f, %f, %f>\n",
       Tv->matrix[2][0], Tv->matrix[2][1], Tv->matrix[2][2], Tv->matrix[2][3]);
printf("  <%f, %f, %f, %f>\n",
       Tv->matrix[3][0], Tv->matrix[3][1], Tv->matrix[3][2], Tv->matrix[3][3]);
*/
   free(Tt);
}

/* mx4_transform: transform 4-vector p by matrix m */
/*  yielding q: q = p*m */
void
mx4_transform(px, py, pz, pw, m, qxp, qyp, qzp, qwp)
   DBL px, py, pz, pw, *qxp, *qyp, *qzp, *qwp;
   MATRIX m;
{
   *qxp = px*m[0][0] + py*m[1][0] + pz*m[2][0] + pw*m[3][0];
   *qyp = px*m[0][1] + py*m[1][1] + pz*m[2][1] + pw*m[3][1];
   *qzp = px*m[0][2] + py*m[1][2] + pz*m[2][2] + pw*m[3][2];
   *qwp = px*m[0][3] + py*m[1][3] + pz*m[2][3] + pw*m[3][3];
}

void
project_points(VIEWPOINT *Eye, int no_of_points, VECTOR *pt,
               DBL *vx, DBL *vy, DBL *bounds)
{
   int i;
   DBL z, w;
   for (i=0; i<no_of_points; i++) {
      mx4_transform(pt[i].x, pt[i].y, pt[i].z, 1.0, WS.matrix,
                    &vx[i], &vy[i], &z, &w);
      vx[i] /= w;
      vy[i] /= w;
      }
}
